(* $Id: pax_tar.ml 774 2008-05-23 14:10:35Z gerd $ *)

(* Reading/writing of tar files *)

(* Infos:
   - https://gps.dynxs.de/private/techdocs/UNIX/single-unix-v2-1997/xcu/pax.html
   - http://www.gnu.org/software/tar/manual/html_section/Standard.html
 *)


open Pax_types

exception Format_error of string


let null_block = String.make 512 '\000'

let is_null_block buf =
  buf = null_block


let read_string name buf pos max_len null_required =
  let n = ref 0 in
  while !n < max_len && buf.[ pos + !n ] <> '\000' do
    incr n
  done;
  if null_required && !n = max_len then raise (Format_error name);
  String.sub buf pos !n


let read_int64 name buf pos max_len =
  (* Old versions of gtar put leading blanks in number fields *)
  let s = ref 0 in
  while !s < max_len && buf.[ pos + !s ] = ' ' do
    incr s
  done;
  let n = ref !s in
  while !n < max_len && buf.[ pos + !n ] >= '0' && buf.[ pos + !n ] <= '7' do
    incr n
  done;
  if !n = max_len then 
    raise (Format_error name);
  if buf.[ pos + !n ] <> '\000' && buf.[ pos + !n ] <> ' ' then 
    raise (Format_error name);
  Int64.of_string ("0o" ^ String.sub buf (pos + !s) (!n - !s))


let read_int name buf pos max_len =
  let n = read_int64 name buf pos max_len in
  Int64.to_int n


let get_cksum buf =
  let s = ref 0 in
  for k = 0 to 511 do
    let c = 
      if k >= 148 && k <= 155 then ' ' else buf.[k] in
    s := (!s + Char.code c) land 0x1ffff
  done;
  !s


let read_metadata buf =
  let magic = read_string "magic" buf 257 8 true in
  let format, is_extended =
    match magic with
      | "ustar" -> "ustar", true
      | "ustar  " -> "oldgnu", true
      | "" -> "tar", false
      | _ -> raise (Format_error "magic") in
  let name_sfx = 
    read_string "name" buf 0 100 false in
  if name_sfx = "" then raise (Format_error "empty name");
  let name_pre = 
    if is_extended then read_string "prefix" buf 345 155 false else "" in
  let name =
    if name_pre = "" then name_sfx else name_pre ^ "/" ^ name_sfx in
  let mode0 =
    read_int "mode" buf 100 8 in
  let mode = mode0 land 0o6777 in
  (* Only standard modes are returned *)
  let uid = read_int "uid" buf 108 8 in
  let uname = 
    if is_extended then Some(read_string "uname" buf 265 32 true) else None in
  let gname = 
    if is_extended then Some(read_string "gname" buf 297 32 true) else None in
  let gid = read_int "gid" buf 116 8 in
  let size = read_int64 "size" buf 124 12 in
  let mtime = read_int64 "mtime" buf 136 12 in
  let cksum = read_int "cksum" buf 148 8 in
  let typeflag = buf.[156] in
  let filetype =
    match typeflag with
      | '\000' | '0' -> 
	  if format = "tar" && name.[ String.length name - 1 ] = '/' then
	    `Directory
	  else
	    `Regular
      | '1' -> 
	  `Hardlink(read_string "linkname" buf 157 100 false)
      | '2' ->
	  `Symlink(read_string "linkname" buf 157 100 false)
      | '5' ->
	  `Directory
      (* The following are only recognized if this is an extended format: *)
      | '3' ->
	  if not is_extended then raise (Format_error "Unexpected char device");
	  `Char(read_int "devmajor" buf 329 8, 
		read_int "devminor" buf 337 8)
      | '4' ->
	  if not is_extended then raise (Format_error "Unexpected block device");
	  `Block(read_int "devmajor" buf 329 8,
		 read_int "devminor" buf 337 8)
      | '6' ->
	  `Fifo
      | _ ->
	  `Other(format, String.make 1 typeflag) in
  let have_data =
    match filetype with
      | `Regular | `Other(_,_) -> true
      | _ -> false in
  (* Finally check the checksum: *)
  let found_cksum = get_cksum buf in
  if found_cksum <> (cksum land 0x1ffff) then raise(Format_error "bad checksum");
  ( object
      method name = name
      method mode = mode
      method size = size
      method owner_uid = Some uid
      method owner_uname = uname
      method owner_gid = Some gid
      method owner_gname = gname
      method mtime = mtime
      method filetype = filetype
      method format = Some format
      method have_data = have_data
    end : metadata
  )


let skip ch =
  let dest = new Netchannels.output_null() in
  dest # output_channel ch


class tar_reader (ch : Netchannels.in_obj_channel) (buf : string) 
                 : archive_rd =
  (* buf: a 512-byte string *)
object(self)
  initializer (
    ignore(self # get_metadata)
  )

  val mutable metadata = None
  val mutable data_read = `None
  val mutable remaining_blocks = 0

  method get_metadata =
    match metadata with
      | Some m -> m
      | None ->
	  ( try 
	      self # read_block();
	      let m = read_metadata buf in
	      metadata <- Some m;
	      m
	    with
	      | Format_error detail ->
		  (* Now check if the block is all zeroes *)
		  if is_null_block buf then (
		    (* If the next block is null, too, this is the end of
                       the archive
		     *)
		    self # read_block();
		    if is_null_block buf then
		      raise End_of_file
		    else
		      failwith ("Cannot read file header: " ^ detail)
		  )
		  else
		    failwith ("Cannot read file header: " ^ detail)
	  )

  method get_data() =
    if data_read <> `None then
      failwith "get_data: already started";
    let m = self # get_metadata in
    let data_blocks =
      if m#have_data then
	Int64.to_int (Int64.shift_right_logical (Int64.add m#size 511L) 9)
      else
	0 in
    remaining_blocks <- data_blocks;
    let remaining_length = ref (if m#have_data then m#size else 0L) in
    let block_len = ref (Int64.to_int (min !remaining_length 512L)) in
    let block_pos = ref !block_len in
    let pos = ref 0 in
    let closed = ref false in
    let data_ch =
      ( object
	  inherit Netchannels.augment_raw_in_channel
	  method input s p l =
	    if !closed then raise Netchannels.Closed_channel;
	    if !block_pos >= !block_len then (
	      if remaining_blocks = 0 then (
		data_read <- `End;
		raise End_of_file;
	      );
	      self # read_block();
	      remaining_blocks <- remaining_blocks - 1;
	      block_pos := 0;
	      block_len := Int64.to_int (min !remaining_length 512L);
	    );
	    let n = min (!block_len - !block_pos) l in
	    String.blit buf !block_pos s p n;
	    block_pos := !block_pos + n;
	    pos := !pos + n;
	    remaining_length := Int64.sub !remaining_length (Int64.of_int n);
	    if !remaining_length = 0L then
	      data_read <- `End;
	    n
	  method close_in() = 
	    closed := true
	  method pos_in = 
	    !pos
	end : Netchannels.in_obj_channel
      ) in
    data_read <- `Progress data_ch;
    `Channel data_ch


  method next() =
    while remaining_blocks > 0 do
      self # read_block();
      remaining_blocks <- remaining_blocks - 1
    done;
    ( try
	let n = new tar_reader ch buf in
	Some n
      with 
	| End_of_file -> None
    )


  
  method private read_block() =
    try
      ch # really_input buf 0 512
    with
      | End_of_file ->
	  failwith "Unexpected end of file"
end

let write_string buf pos field_len null_required s =
  let l = String.length s in
  if l > field_len || (l = field_len && null_required) then
    failwith "String too long";
  String.blit s 0 buf pos l;
  String.fill buf (pos + l) (field_len - l) '\000'
  

let write_int64 buf pos field_len n =
  String.fill buf pos (field_len - 1) '0';
  buf.[pos + field_len - 1] <- '\000';
  let a = ref n in
  let k = ref(field_len - 2) in
  while !a > 0L && !k >= 0 do
    buf.[pos + !k] <- Char.chr(48 + Int64.to_int (Int64.logand !a 7L));
    a := Int64.shift_right_logical !a 3;
    decr k
  done;
  if !a > 0L then
    failwith "Number too big"


let write_int buf pos field_len n =
  write_int64 buf pos field_len (Int64.of_int n)


let write_metadata ~only_metadata buf (m:metadata) =
  (* We always write standard "ustar" headers *)
  ( match m#format with
      | None -> ()
      | Some "ustar" -> ()
      | Some n -> failwith("Cannot write tar headers in this format: " ^ n)
  );
  String.fill buf 0 512 '\000';
  let name = m#name in
  let lname = String.length name in
  if lname <= 100 then (
    write_string buf 0 100 false name;
  )
  else (
    (* Find the first '/' in the last 101 chars of the name. There is the
       corner case that the name starts with a slash, followed by 100 chars.
       In that case, don't break the name up at this slash.
     *)
    try
      let search_at =
	if lname = 101 && name.[0] = '/' then 1 else lname - 101 in
      let slash_pos = String.index_from name search_at '/' in (* or Not_found*)
      if slash_pos - 1 > 155 then
	failwith "Path name does not fit into tar header";
      let name_sfx = String.sub name (slash_pos + 1) (lname - slash_pos - 1) in
      let name_pre = String.sub name 0 (slash_pos - 1) in
      write_string buf 0 100 false name_sfx;
      write_string buf 345 155 false name_pre;
    with
      | Not_found ->
	  failwith "Path name does not fit into tar header"
  );

  write_int buf 100 8 m#mode;

  let uid =
    match m#owner_uid with Some id -> id | None -> 0 in
  let gid =
    match m#owner_gid with Some id -> id | None -> 0 in
  let uname =
    match m#owner_uname with Some n -> n | None -> "" in
  let gname =
    match m#owner_gname with Some n -> n | None -> "" in
  let size = if only_metadata then 0L else m#size in

  write_int buf 108 8 uid;
  write_int buf 116 8 gid;
  write_string buf 265 32 true uname;
  write_string buf 297 32 true gname;
  write_int64 buf 124 12 size;
  write_int64 buf 136 12 m#mtime;
  String.fill buf 148 8 ' ';   (* updated later *)
  write_string buf 257 6 true "ustar";
  buf.[263] <- '0';
  buf.[264] <- '0';
  write_int buf 329 8 0;
  write_int buf 337 8 0;
  ( match m#filetype with
      | `Regular ->
	  buf.[156] <- '0'
      | `Directory ->
	  buf.[156] <- '5'
      | `Hardlink link ->
	  buf.[156] <- '1';
	  ( try
	      write_string buf 157 100 false link
	    with _ ->
	      failwith "Link name does not fit into tar header"
	  )
      | `Symlink link ->
	  buf.[156] <- '2';
	  ( try
	      write_string buf 157 100 false link
	    with _ ->
	      failwith "Link name does not fit into tar header"
	  )
      | `Char(major,minor) ->
	  buf.[156] <- '3';
	  write_int buf 329 8 major;
	  write_int buf 337 8 minor
      | `Block(major,minor) ->
	  buf.[156] <- '4';
	  write_int buf 329 8 major;
	  write_int buf 337 8 minor
      | `Fifo ->
	  buf.[156] <- '6'
      | `Other(_,_) ->
	  failwith "Writing extension headers is not supported"
  );
  (* Checksum: *)
  let cksum = get_cksum buf in
  write_int buf 148 8 cksum


class tar_writer (ch : Netchannels.out_obj_channel) (buf : string) =
object(self)
  val mutable state = `Extensible

  method put_metadata_only (m:metadata) =
    let datach = self # put_gen ~only_metadata:true m in
    datach # close_out()

  method put_data (m:metadata) =
    let only_metadata =
      match m#filetype with
	| `Regular -> false
	| _ -> true in
    self # put_gen ~only_metadata m

  method private put_gen ~only_metadata (m:metadata) =
    if state <> `Extensible then
      failwith "Cannot put more elements into the archive";
    write_metadata ~only_metadata buf m;
    ch # output_string buf;
    
    let announced_size =
      if only_metadata then 0L else m#size in

    state <- `Data;

    let pos_out = ref 0L in
    let block_out = ref 0 in  (* pos_out mod 512 *)

    let datach =
      ( object (chself)
	  inherit Netchannels.augment_raw_out_channel
	  method output s p l =
	    if state <> `Data then
	      failwith "Cannot output data here";
	    let n = ch # output s p l in
	    pos_out := Int64.add !pos_out (Int64.of_int n);
	    block_out := (!block_out + n) land 511;
	    n

	  method pos_out =
	    Int64.to_int !pos_out

	  method flush() =
	    ch # flush();

	  method close_out() =
	    if state <> `Data then
	      failwith "Cannot output data here";
	    if !pos_out <> announced_size then
	      failwith "Actual data length does not match announced length";
	    if !block_out > 0 then
	      ch # output_string (String.make (512 - !block_out) '\000');
	    state <- `Extensible;
	    ch # flush()

	end :Netchannels.out_obj_channel
      ) in

    datach
    

  method close() =
    if state <> `Extensible then
      failwith "Cannot close archive here";
    ch # output_string null_block;
    ch # output_string null_block;
    ch # flush();
    state <- `Closed


end



class tar (file:file) : archive =
object(self) 
  method archive_type = "tar"
  method read_archive() =
    let ch = file # read_file() in
    let buf = String.create 512 in
    ( try Some(new tar_reader ch buf) 
      with End_of_file -> None
    )

  method create_archive() =
    let ch = file # create_file () in
    let buf = String.create 512 in
    new tar_writer ch buf
end

